# -*- coding: utf-8 -*-
"""
/***************************************************************************
 BasemapLoaderPluginDialog
                                 A QGIS plugin
 タイルデータの追加・管理・表示を行うプラグイン
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                             -------------------
        begin                : 2024-10-09
        git sha              : $Format:%H$
        copyright            : (C) 2024 by AeroAsahi Corporation
        email                : plugintest@test
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os

from qgis.PyQt import uic
from qgis.PyQt import QtWidgets
from qgis.PyQt.QtWidgets import (
    QTableWidget,
    QTableWidgetItem
)
from qgis.utils import iface
from qgis.core import (
    QgsApplication,
    Qgis,
    QgsMessageLog,
    QgsProject,
    QgsRasterLayer,
    QgsVectorTileLayer
)
import json
import traceback

# This loads your .ui file so that PyQt can populate your plugin with the elements from Qt Designer
FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'basemap_loader_plugin_dialog_base.ui'))

# JSONファイルの保存先の定義
JSON_FILE = os.path.join(QgsApplication.qgisSettingsDirPath(), "python", "plugins", "basemap_loader_plugin", "basemap.json")
# メッセージのテンプレート
WRITE_ERR_MSG = "背景地図管理ファイルへの書き込みに失敗しました。詳細はログメッセージパネルをご確認ください"
CREATE_ERR_MSG = "背景地図管理ファイルの作成に失敗しました。詳細はログメッセージパネルをご確認ください"

class BasemapLoaderPluginDialog(QtWidgets.QDialog, FORM_CLASS):
    """QGIS用のプラグインダイアログ

    背景地図を追加、管理し、タイルレイヤをQGISのキャンバスに表示するためのダイアログ
    """
        
    def __init__(self, parent=None):
        """コンストラクタ。

        ユーザーインターフェースの初期化や、テーブルウィジェットの設定を行う

        Args:
            parent (QWidget, optional): 親ウィジェット。デフォルトはNone。
        """

        super(BasemapLoaderPluginDialog, self).__init__(parent)
        # Set up the user interface from Designer through FORM_CLASS.
        # After self.setupUi() you can access any designer object by doing
        # self.<objectname>, and you can use autoconnect slots - see
        # http://qt-project.org/doc/qt-4.8/designer-using-a-ui-file.html
        # #widgets-and-dialogs-with-auto-connect
        self.setupUi(self)
        self.iface = iface

        # テーブルウィジェットを行選択にする
        self.tableWidget.setSelectionBehavior(QTableWidget.SelectRows)

        # セルの編集を禁止にする
        self.tableWidget.setEditTriggers(QTableWidget.NoEditTriggers)

        # デフォルトで表示するための背景地図リストのJSONファイルを作成
        if not (os.path.isfile(JSON_FILE)):
            try:
                with open(JSON_FILE, 'w', encoding='utf-8') as f:
                    json.dump(self.create_default_basemap(), f, ensure_ascii=False, indent=4)
            except OSError as e:
                self.show_message("エラー", CREATE_ERR_MSG, Qgis.Warning)
                QgsMessageLog.logMessage(traceback.format_exc(), "プラグイン", QgsMessageLog.WARNING)
                raise

        # イベントハンドラの定義
        self.pushButton_delete.clicked.connect(self.delete_tile_settings)
        self.pushButton_addMap.clicked.connect(self.add_tile_layer)
        self.pushButton_addSettings.clicked.connect(self.save_tile_settings)

        # 背景地図リストのJSONファイルの中身をtableWidgetのアイテムとして追加する
        basemap_data = self.load_json()
        self.add_table_item(self.tableWidget, basemap_data)

    def add_tile_layer(self):
        """選択されたタイルデータをマップキャンバスに追加する

        Raises:
            ValueError: 選択されたデータが不正な場合に発生する
        """
        selected_row = self.tableWidget.currentRow()
        if selected_row >= 0:
            layer = None
            # 選択行の全ての列の値を取得する
            selected_data = self.get_row_values()

            # URIの定義
            uri = f'styleUrl={selected_data[5]}&type=xyz&url={selected_data[1]}&zmax={selected_data[3]}&zmin={selected_data[2]}&http-header:referer={selected_data[4]}'
            # レイヤインスタンスの作成する
            if selected_data[6] == "raster":
                layer = QgsRasterLayer(uri, selected_data[0], 'wms')
            else:
                layer = QgsVectorTileLayer(uri, selected_data[0])

            # マップキャンバスにタイルを追加する
            QgsProject.instance().addMapLayer(layer) 

    def save_tile_settings(self):
        """入力されたタイル設定をJSONファイルに保存し、テーブルに追加する

        Raises:
            ValueError: 入力データが不正な場合に発生する
        """
        if not self.validate_input():
            return
        
        # テーブルの行数を取得する
        row_position = self.tableWidget.rowCount()
        # 新しい行の追加
        self.tableWidget.insertRow(row_position)

        data = {
            "url": self.lineEdit_url.text(),
            "minzl": self.spinBox_minZl.text(),
            "maxzl": self.spinBox_maxZl.text(),
            "referer": self.lineEdit_referer.text(),
            "styleurl": self.lineEdit_styleUrl.text(),
            "type": self.comboBox_type.currentText()
        }

        # 入力された内容をJSONファイルに保存する
        basemap_data = self.load_json()
        basemap_data[self.lineEdit_title.text()] = data
        self.save_json(basemap_data)

        self.add_table_item(self.tableWidget, basemap_data)

        self.show_message("情報", "タイル設定を追加・保存しました", Qgis.Info)

    def delete_tile_settings(self):
        """選択行のタイル設定をJSONファイルから削除する

        Raises:
            ValueError: 選択データが不正な場合に発生する
        """
        # 選択行のインデックスを取得
        selected_row = self.tableWidget.currentRow()
        if selected_row >= 0:
            # 選択行の背景地図タイトルの値を取得
            item = self.tableWidget.item(selected_row, 0)

            # 削除後のJSONをファイルに反映する
            basemap_data = self.load_json()
            basemap_data.pop(item.text())
            self.save_json(basemap_data)

            self.tableWidget.removeRow(selected_row)

            self.show_message("情報", "選択したタイル設定を削除しました", Qgis.Info)

    def get_row_values(self):
        """選択された行の値を取得して返却する

        Returns:
            list: 選択された行の値のリスト
        """
        selected_row = self.tableWidget.currentRow()
        column_count = self.tableWidget.columnCount()

        row_values = []
        for column in range(column_count):
            item = self.tableWidget.item(selected_row, column)
            if item is not None:
                row_values.append(item.text())

        return row_values        

    def add_table_item(self, tableWidget, basemap_data):
        """テーブルウィジェットに背景地図データを追加する

        Args:
            tableWidget (QTableWidget): 背景地図リストを表示するテーブルウィジェットオブジェクト
            basemap_data (dict): 背景地図のJSONデータ
        """
        tableWidget.setRowCount(0)
        row = 0

        for key, value in basemap_data.items():
            tableWidget.insertRow(row)

            tableWidget.setItem(row, 0, QTableWidgetItem(key))
            tableWidget.setItem(row, 1, QTableWidgetItem(value['url']))
            tableWidget.setItem(row, 2, QTableWidgetItem(str(value['minzl'])))
            tableWidget.setItem(row, 3, QTableWidgetItem(str(value['maxzl'])))
            tableWidget.setItem(row, 4, QTableWidgetItem(value['referer']))
            tableWidget.setItem(row, 5, QTableWidgetItem(value['styleurl']))
            tableWidget.setItem(row, 6, QTableWidgetItem(value['type']))

            row += 1

    def show_message(self, message_title, message, level):
        """メッセージバーにメッセージを表示する

        Args:
            message_title (str): メッセージのタイトル
            message (str): メッセージの内容
            level (Qgis.MessageLevel): メッセージのレベル（例: Qgis.Info）
        """
        self.iface.messageBar().pushMessage(
            message_title,
            message,
            level=level,
            duration=5
        )

    def load_json(self):
        """JSONファイルを読み込む

        Returns:
            dict: 読み込まれたJSONデータ

        Raises:
            OSError: ファイル読み込みに失敗した場合
            json.JSONDecodeError: JSON形式が不正な場合
        """
        try:
            with open(JSON_FILE, 'r', encoding='utf-8') as f:
                return json.load(f)
        except OSError as e:
            self.show_message("エラー", "JSONファイルの読み込みに失敗しました。", Qgis.Warning)
            QgsMessageLog.logMessage(traceback.format_exc(), "プラグイン", QgsMessageLog.WARNING)
            raise
        except json.JSONDecodeError as e:
            self.show_message("エラー", "JSONファイルの形式が無効です。", Qgis.Warning)
            QgsMessageLog.logMessage(traceback.format_exc(), "プラグイン", QgsMessageLog.WARNING)
            raise

    def save_json(self, data):
        """JSONファイルを読み込む

        Args:
            data (dict): 保存するJSONデータ
        
        Raises:
            OSError: ファイルの書き込みに失敗した場合
        """
        try:
            with open(JSON_FILE, 'w', encoding='utf-8') as f:
                json.dump(data, f, ensure_ascii=False, indent=4)
        except OSError as e:
            self.show_message("エラー", "JSONファイルの書き込みに失敗しました。", Qgis.Warning)
            QgsMessageLog.logMessage(traceback.format_exc(), "プラグイン", QgsMessageLog.WARNING)
            raise

    def validate_input(self):
        """背景地図タイトルとタイルURLの入力項目のバリデーションチェックを行う

        Returns:
            bool: 入力が空でなければTrue、どちらかが空であればFalseを返却する
        """
        if self.lineEdit_title.text() == "":
            self.show_message("エラー", "背景地図のタイトルを入力してください", Qgis.Warning)
            return False
        if self.lineEdit_url.text() == "":
            self.show_message("エラー", "タイルURLを入力してください", Qgis.Warning)
            return False
        return True

    def create_default_basemap(self):
        """デフォルトで読み込む背景地図リストの辞書データを作成する
        
        Returns:
            dict: デフォルトの背景地図設定の辞書データ
        """

        data = {
            "地理院地図（淡色地図）": {
                "url": "https://cyberjapandata.gsi.go.jp/xyz/pale/{z}/{x}/{y}.png",
                "minzl": 0,
                "maxzl": 18,
                "referer": "",
                "styleurl": "",
                "type": "raster"
            },
            "地理院地図（標準地図）": {
                "url": "https://cyberjapandata.gsi.go.jp/xyz/std/{z}/{x}/{y}.png",
                "minzl": 0,
                "maxzl": 18,
                "referer": "",
                "styleurl": "",
                "type": "raster"
            },
            "地理院地図（白地図）": {
                "url": "https://cyberjapandata.gsi.go.jp/xyz/blank/{z}/{x}/{y}.png",
                "minzl": 5,
                "maxzl": 14,
                "referer": "",
                "styleurl": "",
                "type": "raster"
            },
            "地理院地図（写真）": {
                "url": "https://cyberjapandata.gsi.go.jp/xyz/seamlessphoto/{z}/{x}/{y}.jpg",
                "minzl": 14,
                "maxzl": 18,
                "referer": "",
                "styleurl": "",
                "type": "raster"
            },
            "OpenStreetMap": {
                "url": "https://tile.openstreetmap.org/{z}/{x}/{y}.png",
                "minzl": 0,
                "maxzl": 18,
                "referer": "",
                "styleurl": "",
                "type": "raster"
            },
            "地理院地図（ベクトルタイル）": {
                "url": "https://cyberjapandata.gsi.go.jp/xyz/experimental_bvmap/{z}/{x}/{y}.pbf",
                "minzl": 4,
                "maxzl": 17,
                "referer": "",
                "styleurl": "https://maps.gsi.go.jp/vector/data/pale.json",
                "type": "vector"
            },
        }

        return data